﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Microsoft.SqlServer.TransactSql.ScriptDom;

namespace Parsing_T_SQL_Examples
{
    public static partial class Examples
    {
        internal static void HandleSelectStatement(SelectStatement select, SchemaObjectNameVisitor visitor)
        {
            List<string> ctenames = new List<string>();

            if (select.WithCtesAndXmlNamespaces != null)
            {
                foreach (CommonTableExpression cte in select.WithCtesAndXmlNamespaces.CommonTableExpressions)
                {
                    SchemaObjectNameVisitor localvisitor = new SchemaObjectNameVisitor();
                    cte.AcceptChildren(localvisitor);

                    //if name doesn't have schema and is in list of ctes "seen", ignore:
                    //otherwise, put in visitor:
                    foreach (SchemaObjectName objname in localvisitor.SchemaObjectNameList)
                    {
                        if (
                            objname.SchemaIdentifier == null
                            && ctenames.Contains(objname.BaseIdentifier.Value))
                        {
                        }
                        else
                        {
                            visitor.SchemaObjectNameList.Add(objname);
                        }
                    }

                    //add cte name to list:
                    ctenames.Add(cte.ExpressionName.Value);
                }
            }

            {
                SchemaObjectNameVisitor localvisitor = new SchemaObjectNameVisitor();
                select.QueryExpression.AcceptChildren(localvisitor);

                foreach (SchemaObjectName objname in localvisitor.SchemaObjectNameList)
                {
                    if (
                        objname.SchemaIdentifier == null
                        && ctenames.Contains(objname.BaseIdentifier.Value))
                    {
                    }
                    else
                    {
                        visitor.SchemaObjectNameList.Add(objname);
                    }
                }
            }
        }

        public static bool Example3(string tsqlcode)
        {
            TSqlScript sqlScript = null;

            //parse:
            {
                TSql110Parser parser;
                parser = new TSql110Parser(true);
                System.IO.StringReader txtRdr = new System.IO.StringReader(tsqlcode);

                IList<ParseError> errors;
                TSqlFragment sqlFragment = parser.Parse(txtRdr, out errors);

                sqlScript = (TSqlScript)sqlFragment;
                txtRdr.Dispose();
            }

            //look for Schema Objects => SchemaObjectName
            SchemaObjectNameVisitor schemaobjectnamevisitor = new SchemaObjectNameVisitor();

            foreach (TSqlBatch curbatch in sqlScript.Batches)
            {
                foreach (TSqlStatement curstatement in curbatch.Statements)
                {
                    if (curstatement is CreateViewStatement)
                    {
                        CreateViewStatement createviewstatement = (CreateViewStatement)curstatement;
                        HandleSelectStatement(createviewstatement.SelectStatement, schemaobjectnamevisitor);
                    }
                    else
                    {
                        curstatement.AcceptChildren(schemaobjectnamevisitor);
                    }
                }
            }

            foreach (SchemaObjectName schemaobject in schemaobjectnamevisitor.SchemaObjectNameList)
            {
                if (
                    schemaobject.DatabaseIdentifier == null
                    || schemaobject.SchemaIdentifier == null
                    )
                {
                    return false;
                }
            }

            return true;
        }
    }
}
