﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using Microsoft.SqlServer.TransactSql.ScriptDom;

namespace Parsing_T_SQL_Examples
{
    public static partial class Examples
    {
        internal static List<NamedTableReference> InterrogateJoins(JoinTableReference join)
        {
            List<NamedTableReference> ntreflist = new List<NamedTableReference>();

            if (join is QualifiedJoin)
            {
                QualifiedJoin qualjoin = (QualifiedJoin)join;
                {
                    if (qualjoin.FirstTableReference is NamedTableReference)
                    {
                        NamedTableReference ntref = (NamedTableReference)qualjoin.FirstTableReference;
                        ntreflist.Add(ntref);
                    }

                    if (qualjoin.SecondTableReference is NamedTableReference)
                    {
                        NamedTableReference ntref = (NamedTableReference)qualjoin.FirstTableReference;
                        ntreflist.Add(ntref);
                    }
                    else if (qualjoin.SecondTableReference is JoinTableReference)
                    {
                        JoinTableReference joinref = (JoinTableReference)qualjoin.SecondTableReference;
                        ntreflist.AddRange(InterrogateJoins(joinref));
                    }
                }
            }
            else if (join is UnqualifiedJoin)
            {
                UnqualifiedJoin unqualjoin = (UnqualifiedJoin)join;
                {
                    if (unqualjoin.FirstTableReference is NamedTableReference)
                    {
                        NamedTableReference ntref = (NamedTableReference)unqualjoin.FirstTableReference;
                        ntreflist.Add(ntref);
                    }

                    if (unqualjoin.SecondTableReference is NamedTableReference)
                    {
                        NamedTableReference ntref = (NamedTableReference)unqualjoin.FirstTableReference;
                        ntreflist.Add(ntref);
                    }
                    else if (unqualjoin.SecondTableReference is JoinTableReference)
                    {
                        JoinTableReference joinref = (JoinTableReference)unqualjoin.SecondTableReference;
                        ntreflist.AddRange(InterrogateJoins(joinref));
                    }
                }
            }

            return ntreflist;
        }

        internal static bool InterrogateQuerySpecification_SubQueries(QuerySpecification query, List<string> previousCTEs)
        {
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                //query.Accept(localvisitor);
                query.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }
            
            /*
            foreach (SelectElement element in query.SelectElements)
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                element.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }

            if (query.FromClause != null)
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                query.FromClause.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }

            if (query.WhereClause != null)
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                query.WhereClause.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }

            if (query.OrderByClause != null)
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                query.OrderByClause.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }

            if (query.HavingClause != null)
            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                query.HavingClause.AcceptChildren(localvisitor);

                foreach (QuerySpecification subquery in localvisitor.QuerySpecificationList)
                {
                    if (!InterrogateQuerySpecification(subquery, previousCTEs))
                    {
                        return false;
                    }
                }
            }
            */

            return true;
        }

        internal static bool InterrogateQuerySpecification(QuerySpecification query, List<string> previousCTEs)
        {
            List<SelectStarExpression> selectstars = new List<SelectStarExpression>();

            //check if any SelectStarExpression SelectElements:
            foreach (SelectElement element in query.SelectElements)
            {
                if (element is SelectStarExpression)
                {
                    SelectStarExpression starelement = (SelectStarExpression)element;
                    selectstars.Add(starelement);
                }
            }

            if (selectstars.Count > 0)
            {
                List<NamedTableReference> found_ntrefs = new List<NamedTableReference>();

                //check if data comes from a table or not:
                if (query.FromClause != null)
                {
                    foreach (TableReference tableref in query.FromClause.TableReferences)
                    {
                        //one FROM source, is table:
                        if (tableref is NamedTableReference)
                        {
                            NamedTableReference ntref = (NamedTableReference)tableref;
                            found_ntrefs.Add(ntref);
                        }
                        //multiple sources, FULL/INNER/OUTER join:
                        else if (tableref is JoinTableReference)
                        {
                            JoinTableReference joinref = (JoinTableReference)tableref;
                            found_ntrefs.AddRange(InterrogateJoins(joinref));
                        }
                    }
                }

                //exclude any found CTE references:
                if (found_ntrefs.Count > 0)
                {
                    foreach (NamedTableReference ntref in found_ntrefs.Where(x => x.SchemaObject == null))
                    {
                        if (previousCTEs.Contains(ntref.SchemaObject.BaseIdentifier.Value))
                            found_ntrefs.Remove(ntref);
                    }
                }

                //check aliases:
                if (found_ntrefs.Count > 0)
                {
                    foreach (SelectStarExpression curexp in selectstars)
                    {
                        //if no alias, immediately fails:
                        if (curexp.Qualifier == null)
                        {
                            return false;
                        }

                        //check if uses table alias of named table reference:
                        foreach (NamedTableReference ntref in found_ntrefs.Where(x => x.Alias != null))
                        {
                            if (curexp.Qualifier.Identifiers.Count == 1 && curexp.Qualifier.Identifiers[0] == ntref.Alias)
                            {
                                return false;
                            }
                        }

                        //check if uses table alias of named table reference:
                        foreach (NamedTableReference ntref in found_ntrefs)
                        {
                            string curexp_objectname = curexp.Qualifier.Identifiers[0].Value;
                            string curexp_schemaname = (
                                curexp.Qualifier.Identifiers.Count > 1 
                                ? curexp.Qualifier.Identifiers[1].Value
                                : null);
                            string curexp_dbname = (
                                curexp.Qualifier.Identifiers.Count > 2 
                                ? curexp.Qualifier.Identifiers[2].Value
                                : null);
                            string curexp_servername = (
                                curexp.Qualifier.Identifiers.Count > 3 
                                ? curexp.Qualifier.Identifiers[3].Value
                                : null);

                            string ntref_objectname = ntref.SchemaObject.BaseIdentifier.Value;
                            string ntref_schemaname = (
                                ntref.SchemaObject.SchemaIdentifier != null 
                                ? ntref.SchemaObject.SchemaIdentifier.Value
                                : null);
                            string ntref_dbname = (
                                ntref.SchemaObject.DatabaseIdentifier != null
                                ? ntref.SchemaObject.DatabaseIdentifier.Value
                                : null);
                            string ntref_servername = (
                                ntref.SchemaObject.ServerIdentifier != null
                                ? ntref.SchemaObject.ServerIdentifier.Value
                                : null);

                            if (
                                curexp_objectname == ntref_objectname
                                && curexp_schemaname == ntref_schemaname
                                && curexp_dbname == ntref_dbname
                                && curexp_servername == ntref_servername
                                )
                            {
                                return false;
                            }
                        }
                    }
                }
            }

            //check for subqueries:
            return InterrogateQuerySpecification_SubQueries(query, previousCTEs);
        }

        internal static bool HandleStatementWithCtesAndXmlNamespaces(
            StatementWithCtesAndXmlNamespaces hasCTEstatement
            )
        {
            List<string> ctenames = new List<string>();

            if (hasCTEstatement.WithCtesAndXmlNamespaces != null)
            {
                foreach (CommonTableExpression cte in hasCTEstatement.WithCtesAndXmlNamespaces.CommonTableExpressions)
                {
                    QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                    cte.AcceptChildren(localvisitor);
                    foreach (QuerySpecification queryspec in localvisitor.QuerySpecificationList)
                    {
                        bool result = InterrogateQuerySpecification(queryspec, ctenames);
                        if (!result)
                        {
                            return false;
                        }
                    }

                    //add cte name to list:
                    ctenames.Add(cte.ExpressionName.Value);
                }
            }

            {
                QuerySpecificationVisitor localvisitor = new QuerySpecificationVisitor();
                hasCTEstatement.AcceptChildren(localvisitor);
                foreach (QuerySpecification queryspec in localvisitor.QuerySpecificationList)
                {
                    bool result = InterrogateQuerySpecification(queryspec, ctenames);
                    if (!result)
                    {
                        return false;
                    }
                }
            }

            return true;
        }

        public static bool Example4(string tsqlcode)
        {
            TSqlScript sqlScript = null;

            //parse:
            {
                TSql110Parser parser;
                parser = new TSql110Parser(true);
                System.IO.StringReader txtRdr = new System.IO.StringReader(tsqlcode);

                IList<ParseError> errors;
                TSqlFragment sqlFragment = parser.Parse(txtRdr, out errors);

                sqlScript = (TSqlScript)sqlFragment;
                txtRdr.Dispose();
            }

            foreach (TSqlBatch curbatch in sqlScript.Batches)
            {
                foreach (TSqlStatement curstatement in curbatch.Statements)
                {
                    //look for Statements with CTEs, handle explicitly:

                    if (curstatement is StatementWithCtesAndXmlNamespaces)
                    {
                        StatementWithCtesAndXmlNamespaces hasCTEsStatement = (StatementWithCtesAndXmlNamespaces)curstatement;
                        bool foundval = HandleStatementWithCtesAndXmlNamespaces(hasCTEsStatement);
                        if (!foundval)
                        {
                            return false;
                        }
                    }
                    else
                    {
                        QuerySpecificationVisitor queryspecificationvisitor = new QuerySpecificationVisitor();
                        curstatement.AcceptChildren(queryspecificationvisitor);

                        foreach (QuerySpecification curqueryspec in queryspecificationvisitor.QuerySpecificationList)
                        {
                            bool foundval = InterrogateQuerySpecification(curqueryspec, new List<string>());
                            if (!foundval)
                            {
                                return false;
                            }
                        }
                    }
                }
            }

            return true;
        }

        internal class QuerySpecificationVisitor : TSqlConcreteFragmentVisitor
        {
            public List<QuerySpecification> QuerySpecificationList
            {
                get;
                internal set;
            }

            public QuerySpecificationVisitor()
            {
                QuerySpecificationList = new List<QuerySpecification>();
            }

            public override void ExplicitVisit(QuerySpecification node)
            {
                QuerySpecificationList.Add(node);
            }
        }
    }
}
